import Foundation
import AVFoundation
import LiveKitClient
import WebRTC
import AudioToolbox

class AudioRecorderHandler: NSObject {
    private var audioFile: AVAudioFile?
    private let outputFileURL: URL
    private var _isRecording = false
    private let recordingStateSemaphore = DispatchSemaphore(value: 1)
    private let audioFileAccessQueue = DispatchQueue(label: "com.audioFileAccess")
    private var sampleRate: Double = 48000
    private var channels: Int = 1
    private var capturePost: Bool = false
    var isRecording: Bool {
        get {
            recordingStateSemaphore.wait()
            let state = _isRecording
            recordingStateSemaphore.signal()
            return state
        }
        set {
            recordingStateSemaphore.wait()
            _isRecording = newValue
            recordingStateSemaphore.signal()
        }
    }

    init(filePath: String, capturePost: Bool) {
        self.outputFileURL = URL(fileURLWithPath: filePath.appending(capturePost ? ".1.wav" : ".0.wav"))
        self.capturePost = capturePost
        super.init()
    }

    func startRecording() {
        if capturePost {
            Room.audioProcessingModule.capturePostProcessingDelegate = self
        } else {
            Room.audioProcessingModule.renderPreProcessingDelegate = self
        }
        audioFileAccessQueue.sync { [weak self] in
            guard let self = self, !self.isRecording else { return }
            do {
                guard let audioFormat = AVAudioFormat(standardFormatWithSampleRate: self.sampleRate, channels: AVAudioChannelCount(self.channels)) else {
                    print("Failed to create audio format")
                    return
                }

                self.audioFile = try AVAudioFile(forWriting: self.outputFileURL, settings: audioFormat.settings)
                self.isRecording = true
            } catch {
                print("Failed to initialize audio file: \(error)")
                // Handle error as appropriate
            }
        }
    }

    func stopRecording() {
        if capturePost {
            Room.audioProcessingModule.capturePostProcessingDelegate = nil
        } else {
            Room.audioProcessingModule.renderPreProcessingDelegate = nil
        }
        audioFileAccessQueue.sync { [weak self] in
            self?.isRecording = false
            self?.audioFile = nil
        }
     }

    func process(audioBuffer: RTCAudioBuffer) {
        audioFileAccessQueue.sync {[weak self] in
            guard let weakSelf = self, weakSelf.isRecording, let audioFile = weakSelf.audioFile else {
                print("Failed to process(audioBuffer: RTCAudioBuffer")
                return
            }
            // Convert RTCAudioBuffer to AVAudioPCMBuffer
            guard let pcmBuffer = weakSelf.convertToPCMBuffer(audioBuffer) else {
                print("Failed to convert pcmBuffer")
                return
            }
            // Write the PCM buffer to the audio file
            do {
                try audioFile.write(from: pcmBuffer)
            } catch {
                print("Failed to write audio data to file: \(error)")
            }
        }
    }

    private func convertToPCMBuffer(_ audioBuffer: RTCAudioBuffer) -> AVAudioPCMBuffer? {
        let numFrames = audioBuffer.frames  // The number of frames in your audio data
        let numChannels = audioBuffer.channels  // The number of audio channels (e.g., 2 for stereo)

        guard let audioFormat = AVAudioFormat(standardFormatWithSampleRate: self.sampleRate, channels: AVAudioChannelCount(numChannels)) else {
            print("Failed to create AVAudioFormat")
            return nil
        }

        let frameCapacity = AVAudioFrameCount(numFrames)

        // Create an AVAudioPCMBuffer to hold the audio data
        guard let pcmBuffer = AVAudioPCMBuffer(pcmFormat: audioFormat, frameCapacity: frameCapacity) else {
            print("Failed to create AVAudioPCMBuffer")
            return nil
        }

        // Copy the data from each channel of the RTCAudioBuffer to the AVAudioPCMBuffer
        let channelStride = pcmBuffer.stride

        for channelIndex in 0..<numChannels {
            let audioBufferPointer = audioBuffer.rawBuffer(forChannel: channelIndex)

            guard let targetBufferPointer = pcmBuffer.floatChannelData?[channelIndex] else {
                print("Failed to get target buffer for channel \(channelIndex)")
                return nil
            }
            for frameIndex in 0..<numFrames {
                let sampleValue = audioBufferPointer[frameIndex]
                targetBufferPointer[frameIndex * channelStride] = sampleValue
            }
        }

        // Set the frameLength to indicate how many frames of audio data are in the buffer
        pcmBuffer.frameLength = frameCapacity
        // Your PCM data is now in the AVAudioPCMBuffer and can be used with AVFoundation APIs.
        return pcmBuffer
    }
}

// MARK: - RTCAudioCustomProcessingDelegate
extension AudioRecorderHandler: RTCAudioCustomProcessingDelegate {
    public func audioProcessingInitialize(sampleRate sampleRateHz: Int, channels: Int) {
        self.sampleRate = Double(sampleRateHz)
        self.channels = channels
        print("Initialized with sample rate: \(sampleRateHz) and channels: \(channels)")
    }

    public func audioProcessingProcess(audioBuffer: RTCAudioBuffer) {
        process(audioBuffer: audioBuffer)
    }

    public func audioProcessingRelease() {
        print("Releasing audio processing resources")
    }
}
